"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (babel) {
  var t = babel.types;

  // BABEL-TYPES copypasta; see https://github.com/babel/babel/pull/4886
  /* eslint-disable no-unused-vars */

  function getType(val) {
    if (Array.isArray(val)) {
      return "array";
    } else if (val === null) {
      return "null";
    } else if (val === undefined) {
      return "undefined";
    } else {
      return typeof val;
    }
  }

  function assertEach(callback) {
    function validator(node, key, val) {
      if (!Array.isArray(val)) return;

      for (var i = 0; i < val.length; i++) {
        callback(node, `${key}[${i}]`, val[i]);
      }
    }
    validator.each = callback;
    return validator;
  }

  function assertOneOf() {
    for (var _len = arguments.length, vals = Array(_len), _key = 0; _key < _len; _key++) {
      vals[_key] = arguments[_key];
    }

    function validate(node, key, val) {
      if (vals.indexOf(val) < 0) {
        throw new TypeError(`Property ${key} expected value to be one of ${JSON.stringify(vals)} but got ${JSON.stringify(val)}`);
      }
    }

    validate.oneOf = vals;

    return validate;
  }

  function assertNodeType() {
    for (var _len2 = arguments.length, types = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      types[_key2] = arguments[_key2];
    }

    function validate(node, key, val) {
      var valid = false;

      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = types[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var type = _step.value;

          if (t.is(type, val)) {
            valid = true;
            break;
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      if (!valid) {
        throw new TypeError(`Property ${key} of ${node.type} expected node to be of a type ${JSON.stringify(types)} ` + `but instead got ${JSON.stringify(val && val.type)}`);
      }
    }

    validate.oneOfNodeTypes = types;

    return validate;
  }

  function assertNodeOrValueType() {
    for (var _len3 = arguments.length, types = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
      types[_key3] = arguments[_key3];
    }

    function validate(node, key, val) {
      var valid = false;

      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = types[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var type = _step2.value;

          if (getType(val) === type || t.is(type, val)) {
            valid = true;
            break;
          }
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }

      if (!valid) {
        throw new TypeError(`Property ${key} of ${node.type} expected node to be of a type ${JSON.stringify(types)} ` + `but instead got ${JSON.stringify(val && val.type)}`);
      }
    }

    validate.oneOfNodeOrValueTypes = types;

    return validate;
  }

  function assertValueType(type) {
    function validate(node, key, val) {
      var valid = getType(val) === type;

      if (!valid) {
        throw new TypeError(`Property ${key} expected type of ${type} but got ${getType(val)}`);
      }
    }

    validate.type = type;

    return validate;
  }

  function chain() {
    for (var _len4 = arguments.length, fns = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
      fns[_key4] = arguments[_key4];
    }

    function validate() {
      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = fns[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var fn = _step3.value;

          fn.apply(undefined, arguments);
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }
    }
    validate.chainOf = fns;
    return validate;
  }

  /* eslint-enable no-unused-vars */

  function definePluginType(type) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var inherits = {};
    if (opts.inherits) {
      inherits.visitor = t.VISITOR_KEYS[opts.inherits];
      inherits.builder = t.BUILDER_KEYS[opts.inherits];
      inherits.fields = t.NODE_FIELDS[opts.inherits];
      inherits.aliases = t.ALIAS_KEYS[opts.inherits];
    }

    opts.fields = opts.fields || inherits.fields || {};
    opts.visitor = opts.visitor || inherits.visitor || [];
    opts.aliases = opts.aliases || inherits.aliases || [];
    opts.builder = opts.builder || inherits.builder || opts.visitor || [];

    if (opts.deprecatedAlias) {
      t.DEPRECATED_KEYS[opts.deprecatedAlias] = type;
    }

    // ensure all field keys are represented in `fields`
    var _iteratorNormalCompletion4 = true;
    var _didIteratorError4 = false;
    var _iteratorError4 = undefined;

    try {
      for (var _iterator4 = opts.visitor.concat(opts.builder)[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
        var _key5 = _step4.value;

        opts.fields[_key5] = opts.fields[_key5] || {};
      }
    } catch (err) {
      _didIteratorError4 = true;
      _iteratorError4 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion4 && _iterator4.return) {
          _iterator4.return();
        }
      } finally {
        if (_didIteratorError4) {
          throw _iteratorError4;
        }
      }
    }

    for (var _key6 in opts.fields) {
      var field = opts.fields[_key6];

      if (opts.builder.indexOf(_key6) === -1) {
        field.optional = true;
      }
      if (field.default === undefined) {
        field.default = null;
      } else if (!field.validate) {
        field.validate = assertValueType(getType(field.default));
      }
    }

    t.VISITOR_KEYS[type] = opts.visitor;
    t.BUILDER_KEYS[type] = opts.builder;
    t.NODE_FIELDS[type] = opts.fields;
    t.ALIAS_KEYS[type] = opts.aliases;

    // the below should not be necessary; see https://github.com/babel/babel/pull/4886
    t.TYPES.push(type);

    opts.aliases.forEach(function (alias) {
      t.FLIPPED_ALIAS_KEYS[alias] = t.FLIPPED_ALIAS_KEYS[alias] || [alias];
      t.FLIPPED_ALIAS_KEYS[alias].push(type);

      if (!t.TYPES[alias]) t.TYPES.push(alias);
    });
  }

  // HELPER FUNCTIONS

  // Source mapping tools
  // These can be used to create and place Babel nodes at certain relative
  // positions to other Babel nodes.
  /* eslint-disable no-unused-vars */
  function locateAt(newNode, sourceNode) {
    if (sourceNode) {
      newNode.loc = sourceNode.loc;
      newNode.start = sourceNode.start;
      newNode.end = sourceNode.end;
    }
    return newNode;
  }

  function locateBefore(newNode, sourceNode) {
    if (sourceNode) {
      newNode.loc = Object.assign({}, sourceNode.loc);
      newNode.loc.end = newNode.loc.start;
      newNode.start = sourceNode.start;
      newNode.end = sourceNode.start;
    }
    return newNode;
  }

  function locateAfter(newNode, sourceNode) {
    if (sourceNode) {
      newNode.loc = Object.assign({}, sourceNode.loc);
      newNode.loc.start = newNode.loc.end;
      newNode.start = sourceNode.end;
      newNode.end = sourceNode.end;
    }
    return newNode;
  }

  function nodeAt(sourceNode, type) {
    for (var _len5 = arguments.length, args = Array(_len5 > 2 ? _len5 - 2 : 0), _key7 = 2; _key7 < _len5; _key7++) {
      args[_key7 - 2] = arguments[_key7];
    }

    var newNode = t[type].apply(t, args);
    return locateAt(newNode, sourceNode);
  }

  function nodeBefore(sourceNode, type) {
    for (var _len6 = arguments.length, args = Array(_len6 > 2 ? _len6 - 2 : 0), _key8 = 2; _key8 < _len6; _key8++) {
      args[_key8 - 2] = arguments[_key8];
    }

    var newNode = t[type].apply(t, args);
    return locateBefore(newNode, sourceNode);
  }

  function nodeAfter(sourceNode, type) {
    for (var _len7 = arguments.length, args = Array(_len7 > 2 ? _len7 - 2 : 0), _key9 = 2; _key9 < _len7; _key9++) {
      args[_key9 - 2] = arguments[_key9];
    }

    var newNode = t[type].apply(t, args);
    return locateAfter(newNode, sourceNode);
  }
  /* eslint-enable no-unused-vars */

  var inlinedOperator = {
    looseEq: function looseEq(args) {
      return t.binaryExpression("==", args[0], args[1]);
    },
    looseNotEq: function looseNotEq(args) {
      return t.binaryExpression("!=", args[0], args[1]);
    },
    bitwiseNot: function bitwiseNot(args) {
      return t.unaryExpression("~", args[0]);
    },
    bitwiseAnd: function bitwiseAnd(args) {
      return t.binaryExpression("&", args[0], args[1]);
    },
    bitwiseOr: function bitwiseOr(args) {
      return t.binaryExpression("|", args[0], args[1]);
    },
    bitwiseXor: function bitwiseXor(args) {
      return t.binaryExpression("^", args[0], args[1]);
    },
    bitwiseLeftShift: function bitwiseLeftShift(args) {
      return t.binaryExpression("<<", args[0], args[1]);
    },
    bitwiseRightShift: function bitwiseRightShift(args) {
      return t.binaryExpression(">>", args[0], args[1]);
    },
    bitwiseZeroFillRightShift: function bitwiseZeroFillRightShift(args) {
      return t.binaryExpression(">>>", args[0], args[1]);
    }
  };

  function isFunctionDeclaration(node) {
    return node && (t.is("FunctionDeclaration", node) || node.type === "NamedArrowDeclaration");
  }

  function transformTails(path, allowLoops, getNewNode) {
    path.resync();

    var tailPaths = getTailExpressions(path.get("body"), allowLoops);
    var _iteratorNormalCompletion5 = true;
    var _didIteratorError5 = false;
    var _iteratorError5 = undefined;

    try {
      for (var _iterator5 = tailPaths[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
        var tailPath = _step5.value;

        if (!tailPath) continue;

        if (tailPath.isExpressionStatement()) {
          // TODO: add linting to discourage
          if (tailPath.get("expression").isAssignmentExpression()) {
            tailPath.insertAfter(getNewNode(tailPath.node.expression.left, tailPath));
          } else {
            tailPath.replaceWith(getNewNode(tailPath.node.expression, tailPath));
          }
        } else if (tailPath.isVariableDeclaration()) {
          // TODO: handle declarations.length > 1
          // TODO: add linting to discourage
          tailPath.insertAfter(getNewNode(tailPath.node.declarations[0].id, tailPath));
        } else if (isFunctionDeclaration(tailPath.node)) {
          // Need to transform exprs to statements since this is block context.
          var nextNode = getNewNode(tailPath.node.id, tailPath);
          if (t.isExpression(nextNode)) {
            tailPath.insertAfter(t.expressionStatement(nextNode));
          } else {
            tailPath.insertAfter(nextNode);
          }
        }
      }
    } catch (err) {
      _didIteratorError5 = true;
      _iteratorError5 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion5 && _iterator5.return) {
          _iterator5.return();
        }
      } finally {
        if (_didIteratorError5) {
          throw _iteratorError5;
        }
      }
    }
  }

  function validateComprehensionLoopBody(loopBodyPath) {
    loopBodyPath.traverse({
      AwaitExpression(awaitPath) {
        throw awaitPath.buildCodeFrameError("`await` is not allowed within Comprehensions; " + "instead, await the Comprehension (eg; `y <- [for x of xs: x]`).");
      },
      YieldExpression(yieldPath) {
        throw yieldPath.buildCodeFrameError("`yield` is not allowed within Comprehensions.");
      },
      ReturnStatement(returnPath) {
        throw returnPath.buildCodeFrameError("`return` is not allowed within Comprehensions.");
      }
    });
  }

  function wrapComprehensionInIife(bodyVarId, bodyVarInitializer, loopBody) {
    var fn = t.arrowFunctionExpression([], t.blockStatement([t.variableDeclaration("const", [t.variableDeclarator(bodyVarId, bodyVarInitializer)]), loopBody, t.returnStatement(bodyVarId)]));

    return t.callExpression(fn, []);
  }

  function toBlockStatement(body) {
    if (!t.isBlockStatement(body)) {
      if (!t.isStatement(body)) {
        body = t.expressionStatement(body);
      }
      body = t.blockStatement([body]);
    }
    return body;
  }

  function ensureBlockBody(path) {
    if (!t.isBlockStatement(path.node.body)) {
      path.get("body").replaceWith(t.blockStatement([path.node.body]));
    }
  }

  function toPlainFunction(node) {
    var id = node.id,
        params = node.params,
        body = node.body,
        generator = node.generator,
        async = node.async;


    body = toBlockStatement(body);

    var fn = t.isStatement(node) ? t.functionDeclaration(id, params, body, generator, async) : t.functionExpression(id, params, body, generator, async);

    if (node.returnType) fn.returnType = node.returnType;
    if (node.typeParameters) fn.typeParameters = node.typeParameters;
    return fn;
  }

  function replaceWithPlainFunction(path) {
    path.replaceWith(toPlainFunction(path.node));
  }

  function toArrowFunction(node) {
    var id = node.id,
        params = node.params,
        body = node.body,
        async = node.async;


    if (t.isStatement(node)) {
      var fn = t.arrowFunctionExpression(params, body, async);
      if (node.returnType) fn.returnType = node.returnType;
      if (node.typeParameters) fn.typeParameters = node.typeParameters;
      return t.variableDeclaration("const", [t.variableDeclarator(id, fn)]);
    } else {
      // just throw away the id for now...
      // TODO: think of a way to use it? or outlaw named fat-arrow expressions?
      var _fn = t.arrowFunctionExpression(params, body, async);
      if (node.returnType) _fn.returnType = node.returnType;
      if (node.typeParameters) _fn.typeParameters = node.typeParameters;
      return _fn;
    }
  }

  function replaceWithArrowFunction(path) {
    var id = path.get("id");
    path.replaceWith(toArrowFunction(path.node));
    path.scope.registerBinding("const", id);
  }

  function replaceWithBoundFunction(path) {
    var isStatement = t.isStatement(path.node);

    if (isStatement) {
      replaceWithPlainFunction(path);

      var bound = t.callExpression(t.memberExpression(path.node.id, t.identifier("bind")), [t.thisExpression()]);
      var assignToBound = t.expressionStatement(t.assignmentExpression("=", path.node.id, bound));

      path.insertAfter(assignToBound);
    } else {
      var unbound = toPlainFunction(path.node);
      var _bound = t.callExpression(t.memberExpression(unbound, t.identifier("bind")), [t.thisExpression()]);
      path.replaceWith(_bound);
    }
  }

  function isNamedArrowFunction(node) {
    return typeof node.skinny === "boolean";
  }

  // c/p babel-traverse/src/path/family.js getCompletionRecords
  function getTailExpressions(path, allowLoops) {
    var paths = [];

    var add = function add(_path) {
      if (_path) paths = paths.concat(getTailExpressions(_path, allowLoops));
    };

    if (path.isIfStatement()) {
      add(path.get("consequent"));
      add(path.get("alternate"));
    } else if (path.isDoExpression()) {
      add(path.get("body"));
    } else if (allowLoops && (path.isFor() || path.isWhile())) {
      add(path.get("body"));
    } else if (path.isProgram() || path.isBlockStatement()) {
      add(path.get("body").pop());
    } else if (path.isTryStatement()) {
      add(path.get("block"));
      add(path.get("handler"));
      add(path.get("finalizer"));
    } else {
      paths.push(path);
    }

    return paths;
  }

  function addImplicitReturns(path) {
    transformTails(path, false, function (expr) {
      // Source map: treat the return statement as being at the same
      // location as the implicitly returned expression.
      return nodeAt(expr, "returnStatement", expr);
    });
  }

  function containsSuperCall(path) {
    var hasSuper = false;
    path.traverse({
      Super(superPath) {
        if (superPath.parentPath.isCallExpression()) {
          hasSuper = true;
          superPath.stop();
        }
      }
    });
    return hasSuper;
  }

  function ensureConstructorWithSuper(path, constructorPath) {
    path.resync(); // uhh, just in case?
    var node = path.node;

    // add empty constructor if it wasn't there

    if (!constructorPath) {
      var emptyConstructor = t.classMethod("constructor", t.identifier("constructor"), [], t.blockStatement([]));
      emptyConstructor.skinny = true; // mark for super insertion
      path.get("body").unshiftContainer("body", emptyConstructor);
      constructorPath = path.get("body.body.0.body");
    }

    // add super if it wasn't there (unless defined with curly braces)
    if (node.superClass && constructorPath.parentPath.node.skinny && !containsSuperCall(constructorPath)) {
      var superCall = void 0;
      if (constructorPath.parentPath.node.params.length) {
        var params = constructorPath.parentPath.node.params;
        superCall = t.expressionStatement(t.callExpression(t.super(), params));
      } else {
        var argsUid = path.scope.generateUidIdentifier("args");
        var _params = [t.restElement(argsUid)];
        superCall = t.expressionStatement(t.callExpression(t.super(), [t.spreadElement(argsUid)]));
        constructorPath.parentPath.node.params = _params;
      }
      constructorPath.unshiftContainer("body", superCall);
    }

    return constructorPath;
  }

  function bindMethodsInConstructor(path, constructorPath, methodIds) {
    path.resync(); // uhh, just in case?
    var node = path.node;

    // `this.method = this.method.bind(this);`

    var assignments = methodIds.map(function (methodId) {
      assertOneOf(methodId, ["Identifier", "Expression"]);

      var isComputed = !t.isIdentifier(methodId);
      var thisDotMethod = t.memberExpression(t.thisExpression(), methodId, isComputed);
      var bind = t.callExpression(t.memberExpression(thisDotMethod, t.identifier("bind")), [t.thisExpression()]);
      return t.expressionStatement(t.assignmentExpression("=", thisDotMethod, bind));
    });

    // directly after each instance of super(), insert the thingies there.
    if (node.superClass) {
      constructorPath.traverse({
        Super(superPath) {
          if (!superPath.parentPath.isCallExpression()) return;
          var superStatementPath = superPath.getStatementParent();

          // things get super weird when you return super();
          // TODO: consider trying to handle it
          var enclosingReturn = superPath.findParent(function (p) {
            return p.isReturnStatement() && p.getFunctionParent() === constructorPath.parentPath;
          });
          if (enclosingReturn) throw new Error("Can't use => with `return super()`; try removing `return`.");

          superStatementPath.insertAfter(assignments);
        }
      });
    } else {
      constructorPath.unshiftContainer("body", assignments);
    }
  }

  function bindMethods(path, methodIds) {
    var assignId = void 0,
        inExpression = false;
    if (path.isClassDeclaration()) {
      assignId = path.node.id;
    } else if (path.parentPath.isAssignmentExpression() && path.parentPath.parentPath.isExpressionStatement()) {
      assignId = path.parentPath.node.left;
    } else if (path.parentPath.isVariableDeclarator()) {
      assignId = path.parentPath.node.id;
    } else {
      var id = path.isClass() ? "class" : "obj";
      assignId = path.getStatementParent().scope.generateDeclaredUidIdentifier(id);
      inExpression = true;
    }
    assertOneOf(assignId, ["Identifier", "MemberExpression"]);

    var assignments = methodIds.map(function (methodId) {
      // could be computed, eg `['blah']() => {}`
      assertOneOf(methodId, ["Identifier", "Expression"]);
      var isComputed = !t.isIdentifier(methodId);
      var objDotMethod = t.memberExpression(assignId, methodId, isComputed);
      var bind = t.callExpression(t.memberExpression(objDotMethod, t.identifier("bind")), [assignId]);
      return t.assignmentExpression("=", objDotMethod, bind);
    });

    if (inExpression) {
      path.replaceWith(t.sequenceExpression([t.assignmentExpression("=", assignId, path.node)].concat(_toConsumableArray(assignments), [assignId])));
    } else {
      path.getStatementParent().insertAfter(assignments.map(function (a) {
        return t.expressionStatement(a);
      }));
    }
  }

  function blockToExpression(path) {
    var node = path.node;
    if (t.isBlockStatement(node)) {
      path.canSwapBetweenExpressionAndStatement = function () {
        return true;
      };
      path.replaceExpressionWithStatements(path.node.body);
      return path.node;
    } else if (isFunctionDeclaration(node)) {
      // Special case: the parser promotes a block consisting of a single declaration
      // from ExprStatement(FunctionExpr) to FunctionDeclaration. Convert back to
      // expression here.
      var nextNode = t.clone(node);
      if (nextNode.type === "NamedArrowDeclaration") {
        nextNode.type = "NamedArrowExpression";
      } else {
        nextNode.type = "FunctionExpression";
      }
      return nextNode;
    } else if (t.isExpressionStatement(node)) {
      return node.expression;
    } else {
      return node;
    }
  }

  function checkVariableNotShadowed(path) {
    // ignore top-level; same-level will throw a different error.
    if (!path.scope.parent) return;

    for (var id in path.get("declarations.0").getBindingIdentifiers()) {
      if (path.scope.parent.hasBinding(id)) {
        throw path.buildCodeFrameError(`\`${id}\` is shadowed from a higher scope. ` + `If you want to reassign the variable, use \`now ${id} = ...\`. ` + "If you want to declare a new shadowed \`const\` variable, " + `you must use \`const ${id} = ...\` explicitly.`);
      }
    }
  }

  function shouldParseAsLightScript(file) {
    if (!file || !file.opts || !file.opts.filename) return true;
    var filename = file.opts.filename;
    // HACK: for lightscript-eslint, and possibly others

    if (filename === "unknown") return true;

    // TODO: consider "peeking" at the first line for a shebang or 'use lightscript' directive.
    return (
      // HACK: allow parsing .js test files in this repo.
      // TODO: modify `babel-helper-plugin-test-runner` or something instead
      filename.includes("test/fixtures") || filename.includes(".lsc") || filename.includes(".lsx")
    );
  }

  // eg; 'react', 'lodash/fp', './actions'


  // eg; "React", "PropTypes"


  function initializeStdlib(opts) {
    if (opts.stdlib === false) return false;

    if (typeof opts.stdlib === "object") {
      return Object.assign({}, opts.stdlib.lodash === false ? {} : _stdlib.lodashImports);
    }

    return _stdlib.defaultImports;
  }

  function collectStdlibImport(stdlib, imports, specifier) {
    if (!stdlib) return;

    var importPath = stdlib[specifier];

    if (!imports[importPath]) {
      imports[importPath] = [];
    }

    if (imports[importPath].indexOf(specifier) < 0) {
      imports[importPath].push(specifier);
    }
  }

  function insertStdlibImports(path, imports, useRequire) {
    var declarations = [];
    for (var importPath in imports) {
      var specifierNames = imports[importPath];
      var specifiers = [];

      if (useRequire) {
        // eg; `const { map, uniq } = require('lodash');`
        var _iteratorNormalCompletion6 = true;
        var _didIteratorError6 = false;
        var _iteratorError6 = undefined;

        try {
          for (var _iterator6 = specifierNames[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
            var specifierName = _step6.value;

            var importIdentifier = t.identifier(specifierName);
            specifiers.push(t.objectProperty(importIdentifier, importIdentifier, false, true));
          }
        } catch (err) {
          _didIteratorError6 = true;
          _iteratorError6 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion6 && _iterator6.return) {
              _iterator6.return();
            }
          } finally {
            if (_didIteratorError6) {
              throw _iteratorError6;
            }
          }
        }

        var requirePattern = t.objectPattern(specifiers);
        var requireCall = t.callExpression(t.identifier("require"), [t.stringLiteral(importPath)]);
        var requireStmt = t.variableDeclaration("const", [t.variableDeclarator(requirePattern, requireCall)]);
        declarations.push(requireStmt);
      } else {
        // eg; `import { map, uniq } from 'lodash';`
        var _iteratorNormalCompletion7 = true;
        var _didIteratorError7 = false;
        var _iteratorError7 = undefined;

        try {
          for (var _iterator7 = specifierNames[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
            var _specifierName = _step7.value;

            var _importIdentifier = t.identifier(_specifierName);
            specifiers.push(t.importSpecifier(_importIdentifier, _importIdentifier));
          }
        } catch (err) {
          _didIteratorError7 = true;
          _iteratorError7 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion7 && _iterator7.return) {
              _iterator7.return();
            }
          } finally {
            if (_didIteratorError7) {
              throw _iteratorError7;
            }
          }
        }

        var importDeclaration = t.importDeclaration(specifiers, t.stringLiteral(importPath));
        declarations.push(importDeclaration);
      }
    }
    path.unshiftContainer("body", declarations);
  }

  function generateForInIterator(path, type) {
    var idx = path.node.idx || path.scope.generateUidIdentifier("i");
    var len = path.scope.generateUidIdentifier("len");

    var initDeclarations = [t.variableDeclarator(idx, t.numericLiteral(0))];

    var refId = void 0;
    if (path.get(type).isIdentifier()) {
      refId = path.node[type];
    } else {
      // if the target of iteration is a complex expression,
      // create a reference so it only evaluates once
      var refName = type === "object" ? "obj" : "arr";
      refId = path.scope.generateUidIdentifier(refName);
      initDeclarations.unshift(t.variableDeclarator(refId, path.node[type]));
    }

    var keys = void 0;
    if (type === "object") {
      keys = path.scope.generateUidIdentifier("keys");
      initDeclarations.push(t.variableDeclarator(keys, t.callExpression(t.memberExpression(t.identifier("Object"), t.identifier("keys")), [refId])));
    }

    initDeclarations.push(t.variableDeclarator(len, t.memberExpression(type === "object" ? keys : refId, t.identifier("length"))));

    var init = t.variableDeclaration("let", initDeclarations);
    // _i < _len
    var test = t.binaryExpression("<", idx, len);
    // _i++
    var update = t.updateExpression("++", idx);

    ensureBlockBody(path);
    var innerDeclarations = [];
    if (type === "object") {
      var _key10 = path.node.key || path.scope.generateUidIdentifier("k");
      innerDeclarations.push(t.variableDeclaration("const", [t.variableDeclarator(_key10, t.memberExpression(keys, idx, true))]));

      if (path.node.val) {
        innerDeclarations.push(t.variableDeclaration("const", [t.variableDeclarator(path.node.val, t.memberExpression(refId, _key10, true))]));
      }
    } else {
      if (path.node.elem) {
        innerDeclarations.push(t.variableDeclaration("const", [t.variableDeclarator(path.node.elem, t.memberExpression(refId, idx, true))]));
      }
    }

    if (innerDeclarations.length > 0) {
      path.get("body").unshiftContainer("body", innerDeclarations);
    }

    return t.forStatement(init, test, update, path.node.body);
  }

  function replaceWithSafeCall(path, callExpr) {
    var exprs = [];
    var ref = void 0;

    // If callee is a complex expression, create a ref.
    // XXX: t.isIdentifier() doesn't work here because of the babel-types
    // hacks above.
    if (callExpr.callee.type === "Identifier") {
      ref = callExpr.callee;
    } else {
      ref = path.scope.generateDeclaredUidIdentifier("ref");
      exprs.push(t.assignmentExpression("=", ref, callExpr.callee));
    }

    // Generate actual safecall expr
    // f?(x) -> (typeof f === "function") ? f(x) : null
    exprs.push(t.conditionalExpression(t.binaryExpression("===", t.unaryExpression("typeof", ref), t.stringLiteral("function")), t.callExpression(ref, callExpr.arguments), t.nullLiteral()));

    // Replace, using seqexpr if needed
    path.replaceWith(exprs.length > 1 ? t.sequenceExpression(exprs) : exprs[0]);
  }

  function getInlinedOperatorsEnabled(opts) {
    if (opts.stdlib === false) return false;
    if (typeof opts.stdlib === "object" && opts.stdlib.lightscript === false) return false;
    return true;
  }

  function replaceWithInlinedOperator(path, callee, args) {
    if (callee.type !== "Identifier") return false;
    var name = callee.name;
    if (!inlinedOperator[name] || path.scope.hasBinding(name)) return false;
    path.replaceWith(inlinedOperator[name](args));
  }

  function createMatchTest(casePath, hasPlaceholder, discriminantRef) {
    var testPath = casePath.get("test");
    var testNode = testPath.node;

    if (hasPlaceholder) {
      return testNode;
    } else if (t.isRegExpLiteral(testNode)) {
      // /^regexp/.test(ref)
      // XXX: let consequent see the match results? this could be something like:
      // _ref2 = ref.match(/^regexp/); consequent(_ref2);
      return t.callExpression(t.memberExpression(testNode, t.identifier("test")), [discriminantRef]);
    } else {
      return t.binaryExpression("===", discriminantRef, testNode);
    }
  }

  function createMatchContinuation(matchPath, cases, idx, discriminantRef) {
    if (idx >= cases.length) return t.nullLiteral();
    var c = cases[idx];
    var casePath = matchPath.get(`cases.${idx}`);

    var consequent = c.functional ? t.callExpression(c.consequent, [discriminantRef]) : c.consequent;

    if (c.test.type === "MatchElse") return consequent;

    var hasPlaceholder = false;
    var nestedMatches = 0;
    casePath.traverse({
      // Don't traverse into nested matches; they have different
      // discriminants.
      MatchExpression: {
        enter() {
          nestedMatches++;
        },
        exit() {
          nestedMatches--;
        }
      },
      PlaceholderExpression(phPath) {
        if (nestedMatches === 0) {
          hasPlaceholder = true;
          phPath.replaceWith(discriminantRef);
        }
      }
    });

    return t.conditionalExpression(createMatchTest(casePath, hasPlaceholder, discriminantRef), consequent, createMatchContinuation(matchPath, cases, idx + 1, discriminantRef));
  }

  // TYPE DEFINITIONS
  definePluginType("ForInArrayStatement", {
    visitor: ["idx", "elem", "array", "body"],
    aliases: ["Scopable", "Statement", "For", "BlockParent", "Loop", "ForXStatement", "ForIn"],
    fields: {
      elem: {
        validate: assertNodeType("Identifier"),
        optional: true
      },
      idx: {
        validate: assertNodeType("Identifier"),
        optional: true
      },
      array: {
        validate: assertNodeType("Expression")
      },
      body: {
        validate: assertNodeType("Statement")
      }
    }
  });

  definePluginType("ForInObjectStatement", {
    visitor: ["key", "val", "object", "body"],
    aliases: ["Scopable", "Statement", "For", "BlockParent", "Loop", "ForXStatement", "ForIn"],
    fields: {
      key: {
        validate: assertNodeType("Identifier"),
        optional: true
      },
      val: {
        validate: assertNodeType("Identifier"),
        optional: true
      },
      object: {
        validate: assertNodeType("Expression")
      },
      body: {
        validate: assertNodeType("Statement")
      }
    }
  });

  definePluginType("ArrayComprehension", {
    visitor: ["loop"],
    aliases: ["ArrayExpression", "Expression"],
    fields: {
      loop: {
        validate: assertNodeType("ForStatement")
      }
    }
  });

  definePluginType("ObjectComprehension", {
    visitor: ["loop"],
    aliases: ["ObjectExpression", "Expression"],
    fields: {
      loop: {
        validate: assertNodeType("ForStatement")
      }
    }
  });

  definePluginType("TildeCallExpression", {
    visitor: ["left", "right", "arguments"],
    aliases: ["CallExpression", "Expression"],
    fields: {
      left: {
        validate: assertNodeType("Expression")
      },
      right: {
        validate: assertOneOf("Identifier", "MemberExpression")
      },
      arguments: {
        validate: chain(assertValueType("array"), assertEach(assertNodeType("Expression", "SpreadElement")))
      }
    }
  });

  definePluginType("NamedArrowDeclaration", {
    builder: ["id", "params", "body", "skinny", "async", "generator"],
    visitor: ["id", "params", "body", "returnType", "typeParameters"],
    aliases: ["Scopable", "Function", "BlockParent", "FunctionParent", "Statement", "Pureish", "Declaration", "FunctionDeclaration", "NamedArrowFunction"],
    fields: { // DUP in NamedArrowMemberExpression
      id: {
        validate: assertNodeType("Identifier")
      },
      params: {
        validate: chain(assertValueType("array"), assertEach(assertNodeType("LVal")))
      },
      body: {
        validate: assertNodeType("BlockStatement", "Expression")
      },
      skinny: {
        validate: assertValueType("boolean")
      },
      generator: {
        default: false,
        validate: assertValueType("boolean")
      },
      async: {
        default: false,
        validate: assertValueType("boolean")
      }
    }
  });

  definePluginType("NamedArrowExpression", {
    inherits: "NamedArrowDeclaration",
    aliases: ["Scopable", "Function", "FunctionExpression", "BlockParent", "FunctionParent", "Expression", "Pureish", "NamedArrowFunction"]
  });

  definePluginType("NamedArrowMemberExpression", {
    inherits: "NamedArrowExpression",
    fields: { // c/p from NamedArrowExpression except for `object`
      id: {
        validate: assertNodeType("Identifier")
      },
      object: {
        validate: assertNodeType("Identifier", "MemberExpression")
      },
      params: {
        validate: chain(assertValueType("array"), assertEach(assertNodeType("LVal")))
      },
      body: {
        validate: assertNodeType("BlockStatement", "Expression")
      },
      skinny: {
        validate: assertValueType("boolean")
      },
      generator: {
        default: false,
        validate: assertValueType("boolean")
      },
      async: {
        default: false,
        validate: assertValueType("boolean")
      }
    }
  });

  definePluginType("IfExpression", {
    visitor: ["test", "consequent", "alternate"],
    aliases: ["Expression", "Conditional"],
    fields: {
      test: {
        validate: assertNodeType("Expression")
      },
      consequent: {
        validate: assertNodeType("Expression", "BlockStatement", "ExpressionStatement")
      },
      alternate: {
        optional: true,
        validate: assertNodeType("Expression", "BlockStatement", "ExpressionStatement")
      }
    }
  });

  definePluginType("SafeAwaitExpression", {
    builder: ["argument"],
    visitor: ["argument"],
    aliases: ["AwaitExpression", "Expression", "Terminatorless"],
    fields: {
      argument: {
        validate: assertNodeType("Expression")
      }
    }
  });

  definePluginType("SafeMemberExpression", {
    inherits: "MemberExpression",
    aliases: ["MemberExpression", "Expression", "LVal"]
  });

  definePluginType("ExistentialExpression", {
    builder: ["argument"],
    visitor: ["argument"],
    aliases: ["Expression"],
    fields: {
      argument: {
        validate: assertNodeType("Expression")
      }
    }
  });

  definePluginType("MatchExpression", {
    builder: ["discriminant", "cases"],
    visitor: ["discriminant", "cases"],
    aliases: ["Expression", "Conditional"],
    fields: {
      discriminant: {
        validate: assertNodeType("Expression")
      },
      cases: {
        validate: chain(assertValueType("array"), assertEach(assertNodeType("MatchCase")))
      }
    }
  });

  definePluginType("MatchCase", {
    builder: ["test", "consequent", "functional"],
    visitor: ["test", "consequent"],
    fields: {
      test: {
        validate: assertNodeType("Expression")
      },
      consequent: {
        validate: assertNodeType("Expression")
      }
    }
  });

  definePluginType("MatchElse", {
    aliases: ["Expression"]
  });

  definePluginType("PlaceholderExpression", {
    aliases: ["Expression"]
  });

  // traverse as top-level item so as to run before other babel plugins
  // (and avoid traversing any of their output)
  function Program(path, state) {
    if (!shouldParseAsLightScript(state.file)) return;

    var stdlib = initializeStdlib(state.opts);
    var inlinedOperatorsEnabled = getInlinedOperatorsEnabled(state.opts);
    var useRequire = state.opts.stdlib && state.opts.stdlib.require === true;
    var imports = {};

    path.traverse({

      ForInArrayStatement(path) {
        path.replaceWith(generateForInIterator(path, "array"));
      },

      ForInObjectStatement(path) {
        path.replaceWith(generateForInIterator(path, "object"));
      },

      ArrayComprehension(path) {
        validateComprehensionLoopBody(path.get("loop.body"));

        var id = path.scope.generateUidIdentifier("arr");
        transformTails(path.get("loop"), true, function (expr) {
          return t.callExpression(t.memberExpression(id, t.identifier("push")), [expr]);
        });

        path.replaceWith(wrapComprehensionInIife(id, t.arrayExpression(), path.node.loop));
      },

      ObjectComprehension(path) {
        validateComprehensionLoopBody(path.get("loop.body"));

        var id = path.scope.generateUidIdentifier("obj");
        transformTails(path.get("loop"), true, function (seqExpr, tailPath) {
          // Only SeqExprs of length 2 are valid.
          if (seqExpr.type !== "SequenceExpression" || seqExpr.expressions.length !== 2) {
            throw tailPath.buildCodeFrameError("Object comprehensions must end" + " with a (key, value) pair.");
          }

          var keyExpr = seqExpr.expressions[0];
          var valExpr = seqExpr.expressions[1];

          return t.assignmentExpression("=", t.memberExpression(id, keyExpr, true), valExpr);
        });

        path.replaceWith(wrapComprehensionInIife(id, t.objectExpression([]), path.node.loop));
      },

      TildeCallExpression: {
        // run on exit instead of enter so that SafeMemberExpression
        // can process differently from a wrapping CallExpression
        // eg; `a?.b~c()` -> `a == null ? null : c(a.b)`
        exit(path) {
          var args = [path.node.left].concat(_toConsumableArray(path.node.arguments));
          var callExpr = t.callExpression(path.node.right, args);

          if (inlinedOperatorsEnabled && replaceWithInlinedOperator(path, path.node.right, args)) return;
          if (path.node.safe) {
            replaceWithSafeCall(path, callExpr);
          } else {
            path.replaceWith(callExpr);
          }
        }
      },

      CallExpression: {
        exit(path) {
          if (inlinedOperatorsEnabled && replaceWithInlinedOperator(path, path.node.callee, path.node.arguments)) return;
          if (path.node.safe) {
            replaceWithSafeCall(path, path.node);
          }
        }
      },

      NamedArrowFunction(path) {
        if (path.node.skinny) {
          replaceWithPlainFunction(path);
        } else if (path.node.generator) {
          // there are no arrow-generators in ES6, so can't compile to arrow
          replaceWithBoundFunction(path);
        } else {
          replaceWithArrowFunction(path);
        }
      },

      NamedArrowMemberExpression(path) {
        var object = path.node.object;
        var node = path.node;
        delete node.object;
        node.type = "NamedArrowExpression";

        if (!node.skinny) {
          node.skinny = true; // binding here, don't turn into arrow
          node = t.callExpression(t.memberExpression(node, t.identifier("bind")), [object]);
        }

        path.replaceWith(t.assignmentExpression("=", t.memberExpression(object, path.node.id), node));
      },

      ArrowFunctionExpression(path) {
        if (path.node.skinny) {
          replaceWithPlainFunction(path);
        } else if (path.node.generator) {
          replaceWithBoundFunction(path);
        }
      },

      Method(path) {
        if (isNamedArrowFunction(path.node)) {
          path.node.body = toBlockStatement(path.node.body);
        }
      },

      ClassBody(path) {
        var fatArrows = [],
            fatStaticArrows = [],
            constructorPath = void 0;
        path.node.body.forEach(function (method, i) {
          if (!t.isMethod(method)) return;

          if (method.kind === "constructor") {
            constructorPath = path.get(`body.${i}.body`);
          } else if (method.static && method.skinny === false) {
            fatStaticArrows.push(method.key);
            method.skinny = true; // prevent infinite recursion
          } else if (method.skinny === false) {
            fatArrows.push(method.key);
            method.skinny = true; // prevent infinite recursion
          }
        });

        var maybeAddSuper = path.parentPath.node.superClass && constructorPath;
        if (fatArrows.length || maybeAddSuper) {
          constructorPath = ensureConstructorWithSuper(path.parentPath, constructorPath);
        }

        if (fatArrows.length) {
          bindMethodsInConstructor(path.parentPath, constructorPath, fatArrows);
        }

        if (fatStaticArrows.length) {
          bindMethods(path.parentPath, fatStaticArrows);
        }
      },

      ObjectExpression(path) {
        var fatArrows = [];
        path.node.properties.forEach(function (prop) {
          if (t.isMethod(prop) && prop.skinny === false) {
            fatArrows.push(prop.key);
            // bit ugly, but need a way to ensure we don't double-recurse...
            prop.skinny = true;
          }
        });

        if (fatArrows.length) {
          bindMethods(path, fatArrows);
        }
      },

      Function: {
        exit(path) {
          if (path.node.kind === "constructor" || path.node.kind === "set") return;

          var isVoid = path.node.returnType && t.isVoidTypeAnnotation(path.node.returnType.typeAnnotation);

          if (!isVoid) {
            addImplicitReturns(path);
          }

          // As this is an exit visitor, other LSC transforms have reduced
          // arrows to plain FunctionDeclarations by this point.
          if (path.node.type === "FunctionDeclaration") {
            // somehow this wasn't being done... may signal deeper issues...
            path.getFunctionParent().scope.registerDeclaration(path);
          }
        }
      },

      IfExpression(path) {
        var consequent = blockToExpression(path.get("consequent"));

        var alternate = void 0;
        if (path.node.alternate) {
          alternate = blockToExpression(path.get("alternate"));
        } else {
          alternate = t.nullLiteral();
        }

        path.replaceWith(t.conditionalExpression(path.node.test, consequent, alternate));
      },

      AssignmentExpression(path) {
        if (path.node.operator === "<-" || path.node.operator === "<!-") {
          path.node.operator = "=";
        }

        // TODO: consider enforcing `now` for MemberExpression too
        if (t.isMemberExpression(path.node.left)) return;

        if (path.node.isNowAssign === false) {
          throw path.buildCodeFrameError("Incorrect assignment: to reassign, use `now`; to assign as `const`, put on its own line.");
        }
      },

      SafeAwaitExpression(path) {
        var errId = path.scope.generateUidIdentifier("err");

        var tryCatch = t.tryStatement(t.blockStatement([t.returnStatement(t.awaitExpression(path.node.argument))]), t.catchClause(errId, t.blockStatement([t.returnStatement(errId)])));
        var fn = t.arrowFunctionExpression([], t.blockStatement([tryCatch]), true);
        // TODO: returntype annotation
        var iife = t.callExpression(fn, []);
        var awaitExpr = t.awaitExpression(iife);
        path.replaceWith(awaitExpr);
      },

      SafeMemberExpression(path) {
        // x?.y -> x == null ? x : x.y
        // x?[y] -> x == null ? x : x[y]
        var node = path.node;
        var object = node.object;


        var left = void 0;
        if (object.type === "Identifier" || object.type === "SafeMemberExpression") {
          left = object;
        } else {
          var ref = path.scope.generateDeclaredUidIdentifier("ref");
          node.object = ref;
          left = t.assignmentExpression("=", ref, object);
        }

        var nullCheck = t.binaryExpression("==", left, t.nullLiteral());
        node.type = "MemberExpression";
        path.replaceWith(node);

        // Gather trailing subscripts/calls, which are parent nodes:
        // eg; in `o?.x.y()`, group trailing `.x.y()` into the ternary
        var tail = path;
        while (tail.parentPath) {
          var parent = tail.parentPath;
          var hasChainedParent = parent.isMemberExpression() || parent.isCallExpression() && parent.get("callee") === tail || parent.node.type === "TildeCallExpression" && parent.get("left") === tail;

          if (hasChainedParent) {
            tail = tail.parentPath;
          } else {
            break;
          }
        }

        var ternary = t.conditionalExpression(nullCheck, t.nullLiteral(), tail.node);
        tail.replaceWith(ternary);
      },

      AwaitExpression(path) {
        if (path.get("argument").isArrayExpression() || path.node.argument.type === "ArrayComprehension") {
          var promiseDotAllCall = t.callExpression(t.memberExpression(t.identifier("Promise"), t.identifier("all")), [path.node.argument]);
          path.get("argument").replaceWith(promiseDotAllCall);
        }
      },

      VariableDeclaration(path) {
        // Error on auto-const when shadowing variable
        if (path.node.kind === "const") {
          if (path.node.extra && path.node.extra.implicit === true) {
            checkVariableNotShadowed(path);
          }
        }
      },

      // collect functions to be imported for the stdlib
      ReferencedIdentifier(path) {
        if (stdlib === false) return;

        var node = path.node,
            scope = path.scope;

        if (stdlib[node.name] && !scope.hasBinding(node.name)) {
          collectStdlibImport(stdlib, imports, node.name);
        }
      },

      ExistentialExpression(path) {
        path.replaceWith(t.binaryExpression("!=", path.get("argument").node, t.nullLiteral()));
      },

      MatchExpression(path) {
        var exprs = [];
        var node = path.node;
        var ref = void 0;

        if (node.discriminant.type === "Identifier") {
          ref = node.discriminant;
        } else {
          ref = path.scope.generateDeclaredUidIdentifier("ref");
          exprs.push(t.assignmentExpression("=", ref, node.discriminant));
        }

        exprs.push(createMatchContinuation(path, node.cases, 0, ref));
        path.replaceWith(exprs.length > 1 ? t.sequenceExpression(exprs) : exprs[0]);
      }

    });

    insertStdlibImports(path, imports, useRequire);
  }

  return {
    manipulateOptions(opts, parserOpts, file) {
      if (!shouldParseAsLightScript(file)) return;

      opts.parserOpts = opts.parserOpts || {};
      opts.parserOpts.parser = _babylonLightscript.parse;
      parserOpts.plugins.unshift("lightscript");
      // TODO: allow configuration options to disable these, as they slow down parsing
      parserOpts.plugins.push("jsx", "flow");
    },

    visitor: {
      Program
    }

  };
};

var _babylonLightscript = require("@oigroup/babylon-lightscript");

var _stdlib = require("./stdlib");

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

module.exports = exports["default"];